/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { ICollection, IContext, IScribe, ISequencedOperationMessage, IDeltaService, IDocumentRepository, ICheckpointService } from "@fluidframework/server-services-core";
import { ICheckpointManager } from "./interfaces";
/**
 * MongoDB specific implementation of ICheckpointManager
 */
export declare class CheckpointManager implements ICheckpointManager {
    protected readonly context: IContext;
    private readonly tenantId;
    private readonly documentId;
    private readonly documentRepository;
    private readonly opCollection;
    private readonly deltaService;
    private readonly getDeltasViaAlfred;
    private readonly verifyLastOpPersistence;
    private readonly checkpointService;
    private readonly clientFacadeRetryEnabled;
    constructor(context: IContext, tenantId: string, documentId: string, documentRepository: IDocumentRepository, opCollection: ICollection<ISequencedOperationMessage>, deltaService: IDeltaService, getDeltasViaAlfred: boolean, verifyLastOpPersistence: boolean, checkpointService: ICheckpointService);
    /**
     * Writes the checkpoint information to MongoDB
     */
    write(checkpoint: IScribe, protocolHead: number, pending: ISequencedOperationMessage[], noActiveClients: boolean, globalCheckpointOnly: boolean): Promise<void>;
    /**
     * Removes the checkpoint information from MongoDB
     */
    delete(sequenceNumber: number, lte: boolean): Promise<void>;
}
//# sourceMappingURL=checkpointManager.d.ts.map