"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.ScribeLambda = void 0;
const assert_1 = __importDefault(require("assert"));
const util_1 = require("util");
const protocol_definitions_1 = require("@fluidframework/protocol-definitions");
const server_services_core_1 = require("@fluidframework/server-services-core");
const server_services_telemetry_1 = require("@fluidframework/server-services-telemetry");
const double_ended_queue_1 = __importDefault(require("double-ended-queue"));
const _ = __importStar(require("lodash"));
const utils_1 = require("../utils");
const utils_2 = require("./utils");
class ScribeLambda {
    constructor(context, tenantId, documentId, summaryWriter, pendingMessageReader, checkpointManager, scribe, serviceConfiguration, producer, protocolHandler, protocolHead, messages, scribeSessionMetric, transientTenants, disableTransientTenantFiltering, restartOnCheckpointFailure, kafkaCheckpointOnReprocessingOp) {
        this.context = context;
        this.tenantId = tenantId;
        this.documentId = documentId;
        this.summaryWriter = summaryWriter;
        this.pendingMessageReader = pendingMessageReader;
        this.checkpointManager = checkpointManager;
        this.serviceConfiguration = serviceConfiguration;
        this.producer = producer;
        this.protocolHandler = protocolHandler;
        this.protocolHead = protocolHead;
        this.scribeSessionMetric = scribeSessionMetric;
        this.transientTenants = transientTenants;
        this.disableTransientTenantFiltering = disableTransientTenantFiltering;
        this.restartOnCheckpointFailure = restartOnCheckpointFailure;
        this.kafkaCheckpointOnReprocessingOp = kafkaCheckpointOnReprocessingOp;
        this.pendingCheckpointMessages = new double_ended_queue_1.default();
        // Current sequence/msn of the last processed offset
        this.sequenceNumber = 0;
        this.minSequenceNumber = 0;
        // Indicates whether cache needs to be cleaned after processing a message
        this.clearCache = false;
        // Indicates if the lambda was closed
        this.closed = false;
        // Used to control checkpoint logic
        this.checkpointInfo = {
            lastCheckpointTime: Date.now(),
            rawMessagesSinceCheckpoint: 0,
        };
        // Used to checkpoint if no active clients
        this.noActiveClients = false;
        this.globalCheckpointOnly = false;
        this.lastOffset = scribe.logOffset;
        this.setStateFromCheckpoint(scribe);
        this.pendingMessages = new double_ended_queue_1.default(messages);
    }
    async handler(message) {
        var _a, _b, _c, _d, _e, _f, _g, _h;
        // Skip any log messages we have already processed. Can occur in the case Kafka needed to restart but
        // we had already checkpointed at a given offset.
        if (this.lastOffset !== undefined && message.offset <= this.lastOffset) {
            const reprocessOpsMetric = server_services_telemetry_1.Lumberjack.newLumberMetric(server_services_telemetry_1.LumberEventName.ReprocessOps);
            reprocessOpsMetric.setProperties(Object.assign(Object.assign({}, (0, server_services_telemetry_1.getLumberBaseProperties)(this.documentId, this.tenantId)), { kafkaMessageOffset: message.offset, databaseLastOffset: this.lastOffset }));
            this.updateCheckpointMessages(message);
            try {
                if (this.kafkaCheckpointOnReprocessingOp) {
                    this.context.checkpoint(message, this.restartOnCheckpointFailure);
                }
                reprocessOpsMetric.setProperty("kafkaCheckpointOnReprocessingOp", this.kafkaCheckpointOnReprocessingOp);
                reprocessOpsMetric.success(`Successfully reprocessed repeating ops.`);
            }
            catch (error) {
                reprocessOpsMetric.error(`Error while reprocessing ops.`, error);
            }
            return;
        }
        else if (this.lastOffset === undefined) {
            server_services_telemetry_1.Lumberjack.error(`No value for lastOffset`, (0, server_services_telemetry_1.getLumberBaseProperties)(this.documentId, this.tenantId));
        }
        // if lastOffset is undefined or we have skipped all the previously processed ops,
        // we want to set the offset we store in the database equal to the kafka message offset
        this.lastOffset = message.offset;
        const boxcar = (0, server_services_core_1.extractBoxcar)(message);
        for (const baseMessage of boxcar.contents) {
            if (baseMessage.type === server_services_core_1.SequencedOperationType) {
                const value = baseMessage;
                // Skip messages that were already checkpointed on a prior run.
                if (value.operation.sequenceNumber <= this.sequenceNumber) {
                    continue;
                }
                const lastProtocolHandlerSequenceNumber = (_b = (_a = this.pendingMessages.peekBack()) === null || _a === void 0 ? void 0 : _a.sequenceNumber) !== null && _b !== void 0 ? _b : this.protocolHandler.sequenceNumber;
                // Handles a partial checkpoint case where messages were inserted into DB but checkpointing failed.
                if (value.operation.sequenceNumber <= lastProtocolHandlerSequenceNumber) {
                    continue;
                }
                // Ensure protocol handler sequence numbers are monotonically increasing
                // eslint-disable-next-line @typescript-eslint/restrict-plus-operands
                if (value.operation.sequenceNumber !== lastProtocolHandlerSequenceNumber + 1) {
                    // unexpected sequence number. if a pending message reader is available, ask for those ops
                    if (this.pendingMessageReader !== undefined) {
                        // eslint-disable-next-line @typescript-eslint/restrict-plus-operands
                        const from = lastProtocolHandlerSequenceNumber + 1;
                        const to = value.operation.sequenceNumber - 1;
                        const additionalPendingMessages = await this.pendingMessageReader.readMessages(from, to);
                        for (const additionalPendingMessage of additionalPendingMessages) {
                            this.pendingMessages.push(additionalPendingMessage);
                        }
                    }
                    else {
                        const errorMsg = `Invalid message sequence number.` +
                            `Current message @${value.operation.sequenceNumber}.` +
                            `ProtocolHandler @${lastProtocolHandlerSequenceNumber}`;
                        throw new Error(errorMsg);
                    }
                }
                // Add the message to the list of pending for this document and those that we need
                // to include in the checkpoint
                this.pendingMessages.push(value.operation);
                if (this.serviceConfiguration.scribe.enablePendingCheckpointMessages) {
                    this.pendingCheckpointMessages.push(value);
                }
                // Update the current sequence and min sequence numbers
                const msnChanged = this.minSequenceNumber !== value.operation.minimumSequenceNumber;
                this.sequenceNumber = value.operation.sequenceNumber;
                this.minSequenceNumber = value.operation.minimumSequenceNumber;
                if (msnChanged) {
                    // When the MSN changes we can process up to it to save space
                    this.processFromPending(this.minSequenceNumber);
                }
                this.clearCache = false;
                // skip summarize messages that deli already acked
                if (value.operation.type === protocol_definitions_1.MessageType.Summarize &&
                    !((_c = value.operation.serverMetadata) === null || _c === void 0 ? void 0 : _c.deliAcked)) {
                    // ensure the client is requesting a summary for a state that scribe can achieve
                    // the clients summary state (ref seq num) must be at least as high as scribes (protocolHandler.sequenceNumber)
                    if (!this.summaryWriter.isExternal ||
                        value.operation.referenceSequenceNumber >=
                            this.protocolHandler.sequenceNumber) {
                        // Process up to the summary op ref seq to get the protocol state at the summary op.
                        // Capture state first in case the summary is nacked.
                        const prevState = {
                            protocolState: this.protocolHandler.getProtocolState(),
                            pendingOps: this.pendingMessages.toArray(),
                        };
                        this.processFromPending(value.operation.referenceSequenceNumber);
                        // When external, only process the op if the protocol state advances.
                        // This eliminates the corner case where we have
                        // already captured this summary and are processing this message due to a replay of the stream.
                        if (this.protocolHead < this.protocolHandler.sequenceNumber) {
                            try {
                                const scribeCheckpoint = this.generateScribeCheckpoint(this.lastOffset);
                                const operation = value.operation;
                                const summaryResponse = await this.summaryWriter.writeClientSummary(operation, this.lastClientSummaryHead, scribeCheckpoint, this.pendingCheckpointMessages.toArray());
                                // This block is only executed if the writer is not external. For an external writer,
                                // (e.g., job queue) the responsibility of sending ops to the stream is up to the
                                // external writer.
                                if (!this.summaryWriter.isExternal) {
                                    // On a successful write, send an ack message to clients and a control message to deli.
                                    // Otherwise send a nack and revert the protocol state back to pre summary state.
                                    if (summaryResponse.status) {
                                        await this.sendSummaryAck(summaryResponse.message);
                                        await this.sendSummaryConfirmationMessage(operation.sequenceNumber, true, false);
                                        this.updateProtocolHead(this.protocolHandler.sequenceNumber);
                                        this.updateLastSummarySequenceNumber(this.protocolHandler.sequenceNumber);
                                        const summaryResult = `Client summary success @${value.operation.sequenceNumber}`;
                                        (_d = this.context.log) === null || _d === void 0 ? void 0 : _d.info(summaryResult, {
                                            messageMetaData: {
                                                documentId: this.documentId,
                                                tenantId: this.tenantId,
                                            },
                                        });
                                        server_services_telemetry_1.Lumberjack.info(summaryResult, (0, server_services_telemetry_1.getLumberBaseProperties)(this.documentId, this.tenantId));
                                    }
                                    else {
                                        const nackMessage = summaryResponse.message;
                                        await this.sendSummaryNack(nackMessage);
                                        const errorMsg = `Client summary failure @${value.operation.sequenceNumber}. ` +
                                            `Error: ${nackMessage.message}`;
                                        (_e = this.context.log) === null || _e === void 0 ? void 0 : _e.error(errorMsg, {
                                            messageMetaData: {
                                                documentId: this.documentId,
                                                tenantId: this.tenantId,
                                            },
                                        });
                                        server_services_telemetry_1.Lumberjack.error(errorMsg, (0, server_services_telemetry_1.getLumberBaseProperties)(this.documentId, this.tenantId));
                                        this.revertProtocolState(prevState.protocolState, prevState.pendingOps);
                                    }
                                }
                            }
                            catch (ex) {
                                const errorMsg = `Client summary failure @${value.operation.sequenceNumber}`;
                                (_f = this.context.log) === null || _f === void 0 ? void 0 : _f.error(`${errorMsg} Exception: ${(0, util_1.inspect)(ex)}`);
                                server_services_telemetry_1.Lumberjack.error(errorMsg, (0, server_services_telemetry_1.getLumberBaseProperties)(this.documentId, this.tenantId), ex);
                                this.revertProtocolState(prevState.protocolState, prevState.pendingOps);
                                // If this flag is set, we should ignore any storage specific error and move forward
                                // to process the next message.
                                if (this.serviceConfiguration.scribe.ignoreStorageException) {
                                    await this.sendSummaryNack({
                                        message: "Failed to summarize the document.",
                                        summaryProposal: {
                                            summarySequenceNumber: value.operation.sequenceNumber,
                                        },
                                    });
                                }
                                else {
                                    throw ex;
                                }
                            }
                        }
                    }
                    // eslint-disable-next-line unicorn/prefer-switch
                }
                else if (value.operation.type === protocol_definitions_1.MessageType.NoClient) {
                    (0, assert_1.default)(value.operation.referenceSequenceNumber === value.operation.sequenceNumber, `${value.operation.referenceSequenceNumber} != ${value.operation.sequenceNumber}`);
                    (0, assert_1.default)(value.operation.minimumSequenceNumber === value.operation.sequenceNumber, `${value.operation.minimumSequenceNumber} != ${value.operation.sequenceNumber}`);
                    this.noActiveClients = true;
                    this.globalCheckpointOnly = true;
                    const enableServiceSummaryForTenant = this.disableTransientTenantFiltering ||
                        !this.transientTenants.has(this.tenantId);
                    if (this.serviceConfiguration.scribe.generateServiceSummary &&
                        enableServiceSummaryForTenant) {
                        const operation = value.operation;
                        const scribeCheckpoint = this.generateScribeCheckpoint(this.lastOffset);
                        try {
                            const summaryResponse = await this.summaryWriter.writeServiceSummary(operation, this.protocolHead, scribeCheckpoint, this.pendingCheckpointMessages.toArray());
                            if (summaryResponse) {
                                if (this.serviceConfiguration.scribe.clearCacheAfterServiceSummary) {
                                    this.clearCache = true;
                                }
                                await this.sendSummaryConfirmationMessage(operation.sequenceNumber, false, this.serviceConfiguration.scribe.clearCacheAfterServiceSummary);
                                this.updateLastSummarySequenceNumber(operation.sequenceNumber);
                                // Add service summary handle to validParentSummaries so that SummaryWriter knows it is a valid
                                // alternate parent summary handle. Otherwise only lastClientSummaryHead and latest service summary are accepted.
                                this.updateValidParentSummaries(summaryResponse);
                                const summaryResult = `Service summary success @${operation.sequenceNumber}`;
                                (_g = this.context.log) === null || _g === void 0 ? void 0 : _g.info(summaryResult, {
                                    messageMetaData: {
                                        documentId: this.documentId,
                                        tenantId: this.tenantId,
                                    },
                                });
                                server_services_telemetry_1.Lumberjack.info(summaryResult, (0, server_services_telemetry_1.getLumberBaseProperties)(this.documentId, this.tenantId));
                            }
                        }
                        catch (ex) {
                            const errorMsg = `Service summary failure @${operation.sequenceNumber}`;
                            // If this flag is set, we should ignore any storage speciic error and move forward
                            // to process the next message.
                            if (this.serviceConfiguration.scribe.ignoreStorageException) {
                                (_h = this.context.log) === null || _h === void 0 ? void 0 : _h.error(errorMsg, {
                                    messageMetaData: {
                                        documentId: this.documentId,
                                        tenantId: this.tenantId,
                                    },
                                });
                                server_services_telemetry_1.Lumberjack.error(errorMsg, (0, server_services_telemetry_1.getLumberBaseProperties)(this.documentId, this.tenantId), ex);
                            }
                            else {
                                throw ex;
                            }
                        }
                    }
                }
                else if (value.operation.type === protocol_definitions_1.MessageType.SummaryAck) {
                    const operation = value.operation;
                    const content = operation.data
                        ? JSON.parse(operation.data)
                        : operation.contents;
                    this.lastClientSummaryHead = content.handle;
                    // Similar to lastClientSummaryHead, only reset validParentSummaries to undefined
                    // once a new official client summary ack is receieved.
                    // It will be updated to an array if/when summary handles are added.
                    this.validParentSummaries = undefined;
                    // An external summary writer can only update the protocolHead when the ack is sequenced
                    // back to the stream.
                    if (this.summaryWriter.isExternal) {
                        this.updateProtocolHead(content.summaryProposal.summarySequenceNumber);
                        this.updateLastSummarySequenceNumber(content.summaryProposal.summarySequenceNumber);
                    }
                }
                else if (value.operation.type === protocol_definitions_1.MessageType.ClientJoin) {
                    this.globalCheckpointOnly = false;
                }
            }
        }
        // Create the checkpoint
        this.updateCheckpointMessages(message);
        this.checkpointInfo.rawMessagesSinceCheckpoint++;
        if (this.noActiveClients) {
            this.prepareCheckpoint(message, utils_1.CheckpointReason.NoClients);
            this.noActiveClients = false;
        }
        else {
            this.globalCheckpointOnly = false;
            const checkpointReason = this.getCheckpointReason();
            if (checkpointReason !== undefined) {
                // checkpoint the current up-to-date state
                this.prepareCheckpoint(message, checkpointReason);
            }
            else {
                this.updateCheckpointIdleTimer();
            }
        }
    }
    prepareCheckpoint(message, checkpointReason) {
        var _a, _b, _c, _d;
        // Get checkpoint context
        const checkpoint = this.generateScribeCheckpoint(message.offset);
        this.updateCheckpointMessages(message);
        // write the checkpoint with the current up-to-date state
        this.checkpoint(checkpointReason);
        this.checkpointCore(checkpoint, message, this.clearCache);
        this.lastOffset = message.offset;
        const reason = utils_1.CheckpointReason[checkpointReason];
        const checkpointResult = `Writing checkpoint. Reason: ${reason}`;
        const lumberjackProperties = Object.assign(Object.assign({}, (0, server_services_telemetry_1.getLumberBaseProperties)(this.documentId, this.tenantId)), { checkpointReason: reason, lastOffset: this.lastOffset, scribeCheckpointOffset: (_a = this.checkpointInfo.currentCheckpointMessage) === null || _a === void 0 ? void 0 : _a.offset, scribeCheckpointPartition: (_b = this.checkpointInfo.currentCheckpointMessage) === null || _b === void 0 ? void 0 : _b.partition, kafkaCheckpointOffset: (_c = this.checkpointInfo.currentKafkaCheckpointMessage) === null || _c === void 0 ? void 0 : _c.offset, kafkaCheckpointPartition: (_d = this.checkpointInfo.currentKafkaCheckpointMessage) === null || _d === void 0 ? void 0 : _d.partition });
        server_services_telemetry_1.Lumberjack.info(checkpointResult, lumberjackProperties);
    }
    close(closeType) {
        this.logScribeSessionMetrics(closeType);
        this.closed = true;
        this.protocolHandler.close();
    }
    logScribeSessionMetrics(closeType) {
        var _a;
        if ((_a = this.scribeSessionMetric) === null || _a === void 0 ? void 0 : _a.isCompleted()) {
            server_services_telemetry_1.Lumberjack.info("Scribe session metric already completed. Creating a new one.", (0, server_services_telemetry_1.getLumberBaseProperties)(this.documentId, this.tenantId));
            this.scribeSessionMetric = (0, utils_1.createSessionMetric)(this.tenantId, this.documentId, server_services_telemetry_1.LumberEventName.ScribeSessionResult, this.serviceConfiguration);
        }
        (0, utils_1.logCommonSessionEndMetrics)(this.context, closeType, this.scribeSessionMetric, this.sequenceNumber, this.protocolHead, undefined);
    }
    // Advances the protocol state up to 'target' sequence number. Having an exception while running this code
    // is crucial and the document is essentially corrupted at this point. We should start logging this and
    // have a better understanding of all failure modes.
    processFromPending(target) {
        var _a;
        while (this.pendingMessages.length > 0 &&
            // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
            this.pendingMessages.peekFront().sequenceNumber <= target) {
            // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
            const message = this.pendingMessages.shift();
            try {
                if (message.contents &&
                    typeof message.contents === "string" &&
                    message.type !== protocol_definitions_1.MessageType.ClientLeave) {
                    const clonedMessage = _.cloneDeep(message);
                    clonedMessage.contents = JSON.parse(clonedMessage.contents);
                    this.protocolHandler.processMessage(clonedMessage, false);
                }
                else {
                    this.protocolHandler.processMessage(message, false);
                }
            }
            catch (error) {
                (_a = this.context.log) === null || _a === void 0 ? void 0 : _a.error(`Protocol error ${error}`, {
                    messageMetaData: {
                        documentId: this.documentId,
                        tenantId: this.tenantId,
                    },
                });
                server_services_telemetry_1.Lumberjack.error(`Protocol error`, (0, server_services_telemetry_1.getLumberBaseProperties)(this.documentId, this.tenantId), error);
                throw new Error(`Protocol error ${error} for ${this.documentId} ${this.tenantId}`);
            }
        }
    }
    revertProtocolState(protocolState, pendingOps) {
        this.protocolHandler = (0, utils_2.initializeProtocol)(protocolState);
        this.pendingMessages = new double_ended_queue_1.default(pendingOps);
    }
    generateScribeCheckpoint(logOffset) {
        const protocolState = this.protocolHandler.getProtocolState();
        const checkpoint = {
            lastSummarySequenceNumber: this.lastSummarySequenceNumber,
            lastClientSummaryHead: this.lastClientSummaryHead,
            logOffset,
            minimumSequenceNumber: this.minSequenceNumber,
            protocolState,
            sequenceNumber: this.sequenceNumber,
            validParentSummaries: this.validParentSummaries,
        };
        return checkpoint;
    }
    checkpointCore(checkpoint, queuedMessage, clearCache) {
        if (this.closed) {
            return;
        }
        if (this.pendingP) {
            this.pendingCheckpointScribe = checkpoint;
            this.pendingCheckpointOffset = queuedMessage;
            return;
        }
        this.pendingP = clearCache
            ? this.checkpointManager.delete(this.protocolHead, true)
            : this.writeCheckpoint(checkpoint);
        this.pendingP
            .then(() => {
            this.pendingP = undefined;
            this.context.checkpoint(queuedMessage, this.restartOnCheckpointFailure);
            const pendingScribe = this.pendingCheckpointScribe;
            const pendingOffset = this.pendingCheckpointOffset;
            if (pendingScribe && pendingOffset) {
                this.pendingCheckpointScribe = undefined;
                this.pendingCheckpointOffset = undefined;
                this.checkpointCore(pendingScribe, pendingOffset, clearCache);
            }
        })
            .catch((error) => {
            const message = "Checkpoint error";
            server_services_telemetry_1.Lumberjack.error(message, (0, server_services_telemetry_1.getLumberBaseProperties)(this.documentId, this.tenantId), error);
        });
    }
    async writeCheckpoint(checkpoint) {
        const inserts = this.pendingCheckpointMessages.toArray();
        await this.checkpointManager.write(checkpoint, this.protocolHead, inserts, this.noActiveClients, this.globalCheckpointOnly);
        if (inserts.length > 0) {
            // Since we are storing logTails with every summary, we need to make sure that messages are either in DB
            // or in memory. In other words, we can only remove messages from memory once there is a copy in the DB
            const lastInsertedSeqNumber = inserts[inserts.length - 1].operation.sequenceNumber;
            while (this.pendingCheckpointMessages.length > 0 &&
                // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
                this.pendingCheckpointMessages.peekFront().operation.sequenceNumber <=
                    lastInsertedSeqNumber) {
                this.pendingCheckpointMessages.removeFront();
            }
        }
    }
    /**
     * Protocol head is the sequence number of the last summary
     * This method updates the protocol head to the new summary sequence number
     * @param protocolHead - The sequence number of the new summary
     */
    updateProtocolHead(protocolHead) {
        this.protocolHead = protocolHead;
    }
    /**
     * lastSummarySequenceNumber tracks the sequence number that was part of the latest summary
     * This method updates it to the sequence number that was part of the latest summary
     * @param summarySequenceNumber - The sequence number of the operation that was part of the latest summary
     */
    updateLastSummarySequenceNumber(summarySequenceNumber) {
        this.lastSummarySequenceNumber = summarySequenceNumber;
    }
    /**
     * validParentSummaries tracks summary handles for service summaries that have been written since the latest client summary.
     * @param summaryHandle - The handle for a service summary that occurred after latest client summary.
     */
    updateValidParentSummaries(summaryHandle) {
        if (this.validParentSummaries === undefined) {
            this.validParentSummaries = [];
        }
        this.validParentSummaries.push(summaryHandle);
    }
    async sendSummaryAck(contents) {
        const operation = {
            clientSequenceNumber: -1,
            contents,
            data: JSON.stringify(contents),
            referenceSequenceNumber: -1,
            traces: this.serviceConfiguration.enableTraces ? [] : undefined,
            type: protocol_definitions_1.MessageType.SummaryAck,
        };
        return (0, utils_2.sendToDeli)(this.tenantId, this.documentId, this.producer, operation);
    }
    async sendSummaryNack(contents) {
        const operation = {
            clientSequenceNumber: -1,
            contents,
            data: JSON.stringify(contents),
            referenceSequenceNumber: -1,
            traces: this.serviceConfiguration.enableTraces ? [] : undefined,
            type: protocol_definitions_1.MessageType.SummaryNack,
        };
        return (0, utils_2.sendToDeli)(this.tenantId, this.documentId, this.producer, operation);
    }
    // Sends a confirmation back to deli as a signal to update its DSN. Note that 'durableSequenceNumber (dsn)'
    // runs ahead of last summary sequence number (protocolHead). The purpose of dsn is to inform deli about permanent
    // storage so that it can hydrate its state after a failure. The client's are still reponsible for fetching ops
    // from protocolHead to dsn.
    async sendSummaryConfirmationMessage(durableSequenceNumber, isClientSummary, clearCache) {
        const controlMessage = {
            type: server_services_core_1.ControlMessageType.UpdateDSN,
            contents: {
                durableSequenceNumber,
                isClientSummary,
                clearCache,
            },
        };
        const operation = {
            clientSequenceNumber: -1,
            contents: null,
            data: JSON.stringify(controlMessage),
            referenceSequenceNumber: -1,
            traces: this.serviceConfiguration.enableTraces ? [] : undefined,
            type: protocol_definitions_1.MessageType.Control,
        };
        return (0, utils_2.sendToDeli)(this.tenantId, this.documentId, this.producer, operation);
    }
    setStateFromCheckpoint(scribe) {
        this.sequenceNumber = scribe.sequenceNumber;
        this.minSequenceNumber = scribe.minimumSequenceNumber;
        this.lastClientSummaryHead = scribe.lastClientSummaryHead;
        this.lastSummarySequenceNumber = scribe.lastSummarySequenceNumber;
        this.validParentSummaries = scribe.validParentSummaries;
    }
    updateCheckpointMessages(message) {
        // updates scribe checkpoint message
        this.checkpointInfo.currentCheckpointMessage = message;
        if (this.noActiveClients) {
            this.checkpointInfo.nextKafkaCheckpointMessage = undefined;
            this.checkpointInfo.currentKafkaCheckpointMessage = message;
        }
        else {
            // Keeps Kafka checkpoint behind by 1 message (is this necessary?)
            const kafkaCheckpointMessage = this.checkpointInfo.nextKafkaCheckpointMessage;
            this.checkpointInfo.nextKafkaCheckpointMessage = message;
            this.checkpointInfo.currentKafkaCheckpointMessage = kafkaCheckpointMessage;
        }
    }
    // Determines checkpoint reason based on some Heuristics
    getCheckpointReason() {
        const checkpointHeuristics = this.serviceConfiguration.scribe.checkpointHeuristics;
        if (!checkpointHeuristics.enable) {
            // always checkpoint since heuristics are disabled
            return utils_1.CheckpointReason.EveryMessage;
        }
        if (this.checkpointInfo.rawMessagesSinceCheckpoint >= checkpointHeuristics.maxMessages) {
            // exceeded max messages since last checkpoint
            return utils_1.CheckpointReason.MaxMessages;
        }
        if (Date.now() - this.checkpointInfo.lastCheckpointTime >= checkpointHeuristics.maxTime) {
            // exceeded max time since last checkpoint
            return utils_1.CheckpointReason.MaxTime;
        }
        return undefined;
    }
    clearCheckpointIdleTimer() {
        if (this.checkpointInfo.idleTimer !== undefined) {
            clearTimeout(this.checkpointInfo.idleTimer);
            this.checkpointInfo.idleTimer = undefined;
        }
    }
    checkpoint(reason) {
        this.clearCheckpointIdleTimer();
        this.checkpointInfo.lastCheckpointTime = Date.now();
        this.checkpointInfo.rawMessagesSinceCheckpoint = 0;
    }
    updateCheckpointIdleTimer() {
        this.clearCheckpointIdleTimer();
        const initialScribeCheckpointMessage = this.checkpointInfo.currentCheckpointMessage;
        this.checkpointInfo.idleTimer = setTimeout(() => {
            var _a, _b, _c, _d;
            this.checkpointInfo.idleTimer = undefined;
            // verify that the current scribe message matches the message that started this timer
            // same implementation as in Deli
            if (initialScribeCheckpointMessage === this.checkpointInfo.currentCheckpointMessage) {
                this.checkpoint(utils_1.CheckpointReason.IdleTime);
                if (initialScribeCheckpointMessage) {
                    this.checkpointCore(this.generateScribeCheckpoint(initialScribeCheckpointMessage.offset), initialScribeCheckpointMessage, this.clearCache);
                    const checkpointResult = `Writing checkpoint. Reason: IdleTime`;
                    const lumberjackProperties = Object.assign(Object.assign({}, (0, server_services_telemetry_1.getLumberBaseProperties)(this.documentId, this.tenantId)), { checkpointReason: "IdleTime", lastOffset: initialScribeCheckpointMessage.offset, deliCheckpointOffset: (_a = this.checkpointInfo.currentCheckpointMessage) === null || _a === void 0 ? void 0 : _a.offset, deliCheckpointPartition: (_b = this.checkpointInfo.currentCheckpointMessage) === null || _b === void 0 ? void 0 : _b.partition, kafkaCheckpointOffset: (_c = this.checkpointInfo.currentKafkaCheckpointMessage) === null || _c === void 0 ? void 0 : _c.offset, kafkaCheckpointPartition: (_d = this.checkpointInfo.currentKafkaCheckpointMessage) === null || _d === void 0 ? void 0 : _d.partition });
                    server_services_telemetry_1.Lumberjack.info(checkpointResult, lumberjackProperties);
                }
            }
        }, this.serviceConfiguration.scribe.checkpointHeuristics.idleTime);
    }
}
exports.ScribeLambda = ScribeLambda;
//# sourceMappingURL=lambda.js.map