"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.ConnectionCountLogger = void 0;
const server_services_telemetry_1 = require("@fluidframework/server-services-telemetry");
class ConnectionCountLogger {
    constructor(nodeName = "nodeName", cache) {
        this.nodeName = nodeName;
        this.cache = cache;
        this.perClusterKeyName = `totalConnections`;
        this.perNodeKeyName = `totalConnections_${this.nodeName}`;
    }
    incrementConnectionCount() {
        if (!this.cache || !this.cache.incr) {
            return;
        }
        const connectionCountPerNodeMetric = server_services_telemetry_1.Lumberjack.newLumberMetric(server_services_telemetry_1.LumberEventName.ConnectionCountPerNode);
        const totalConnectionCountMetric = server_services_telemetry_1.Lumberjack.newLumberMetric(server_services_telemetry_1.LumberEventName.TotalConnectionCount);
        this.cache
            .incr(this.perNodeKeyName)
            .then((val) => {
            connectionCountPerNodeMetric.setProperty("TotalConnectionCount", val);
            connectionCountPerNodeMetric.success("Connection count incremented for node.");
        })
            .catch((error) => {
            connectionCountPerNodeMetric.error(`Error while incrementing connection count for node.`, error);
        });
        this.cache
            .incr(this.perClusterKeyName)
            .then((val) => {
            totalConnectionCountMetric.setProperty("TotalConnectionCount", val);
            totalConnectionCountMetric.success("Total connection count incremented.");
        })
            .catch((error) => {
            totalConnectionCountMetric.error(`Error while incrementing total connection count for cluster.`, error);
        });
    }
    decrementConnectionCount() {
        if (!this.cache || !this.cache.decr) {
            return;
        }
        const connectionCountPerNodeMetric = server_services_telemetry_1.Lumberjack.newLumberMetric(server_services_telemetry_1.LumberEventName.ConnectionCountPerNode);
        const totalConnectionCountMetric = server_services_telemetry_1.Lumberjack.newLumberMetric(server_services_telemetry_1.LumberEventName.TotalConnectionCount);
        this.cache
            .decr(this.perNodeKeyName)
            .then((val) => {
            connectionCountPerNodeMetric.setProperty("TotalConnectionCount", val);
            connectionCountPerNodeMetric.success("Connection count decremented for node.");
        })
            .catch((error) => {
            connectionCountPerNodeMetric.error(`Error while decrementing connection count for node`, error);
        });
        this.cache
            .decr(this.perClusterKeyName)
            .then((val) => {
            totalConnectionCountMetric.setProperty("TotalConnectionCount", val);
            totalConnectionCountMetric.success("Total connection count decremented.");
        })
            .catch((error) => {
            totalConnectionCountMetric.error(`Error while decrementing total connection count for cluster.`, error);
        });
    }
}
exports.ConnectionCountLogger = ConnectionCountLogger;
//# sourceMappingURL=connectionCountLogger.js.map